/*
 * SPDX-FileCopyrightText: Copyright 2024 Ariadne Conill <ariadne@ariadne.space>
 * SPDX-License-Identifier: ISC
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <sys/xattr.h>

static bool debug = false;
#define DPRINTF(...) \
	if (debug) { fprintf(stderr, "paxrelabel: " __VA_ARGS__); }
#define EPRINTF(...) \
	fprintf(stderr, "paxrelabel: error: " __VA_ARGS__);

static bool
relabel_file(const char *flags, size_t flags_len, const char *target, size_t *relabeled_count)
{
	DPRINTF("relabel %s with flags %s\n", target, flags);

	if (setxattr(target, "user.pax.flags", flags, flags_len, 0) == -1)
	{
		if (errno != ENOENT)
		{
			err(EXIT_FAILURE, "error while relabeling %s with flags %s", target, flags);
			return false;
		}
	}

	*relabeled_count += 1;
	return true;
}

/*
 * Lines beginning with a '#' are comments, otherwise we expect the line
 * to begin with the PaX flags (pemrs) and the binary name.
 */
static bool
process_line(const char *linebuf, const char *file, size_t *relabeled_count)
{
	char valid_flags[] = "pemrs";

	if (linebuf == NULL)
		return true;

	if (*linebuf == '#' || *linebuf == '\n' || !*linebuf)
		return true;

	const char *flags = linebuf + strspn(linebuf, " \t");
	const char *split = flags + strcspn(flags, " \t");
	const char *flag;

	for (flag = flags; flag < split; flag++)
	{
		char *slot = strchr(valid_flags, tolower(*flag));

		if (slot != NULL)
			*slot = *flag;
		else
		{
			EPRINTF("%s: invalid flag %c, only [%s] allowed\n", file, *flag, valid_flags);
			errno = EINVAL;
			return false;
		}
	}

	const char *target = split + strspn(split, " \t");
	const char *target_end = target + strcspn(target, " \t\n");

	size_t targetlen = target_end - target;
	char *targetbuf = alloca(targetlen + 1);
	strlcpy(targetbuf, target, targetlen + 1);

	return relabel_file(valid_flags, (sizeof valid_flags) - 1, targetbuf, relabeled_count);
}

static bool
process_file(int pathfd, const char *file, size_t *relabeled_count)
{
	bool ret = true;

	int fd = openat(pathfd, file, O_RDONLY);
	if (fd < 0)
		return false;

	FILE *f = fdopen(fd, "r");
	if (f == NULL)
	{
		close(fd);
		return false;
	}

	char *linebuf = NULL;
	size_t bufsize = 0;
	while (getline(&linebuf, &bufsize, f) != -1)
	{
		if (!process_line(linebuf, file, relabeled_count))
		{
			ret = false;
			break;
		}
	}

	free(linebuf);
	fclose(f);

	return ret;
}

static bool
process_directory(const char *path, size_t *relabeled_count)
{
	bool ret = true;

	DPRINTF("processing %s\n", path);

	DIR *dir = opendir(path);

	/* if the directory doesn't exist, it is not a fatal error */
	if (dir == NULL)
		return ret;

	int pathfd = dirfd(dir);
	if (pathfd < 0)
		return ret;

	struct dirent *dentry;
	for (dentry = readdir(dir); dentry != NULL; dentry = readdir(dir))
	{
		if (dentry->d_type != DT_REG)
			continue;

		if (!process_file(pathfd, dentry->d_name, relabeled_count))
		{
			ret = false;
			break;
		}
	}

	closedir(dir);

	return ret;
}

int
main(int argc, const char *argv[])
{
	(void) argc;
	(void) argv;
	char search_path[] = PAXRELABEL_SEARCH_PATH;
	char *search_saveptr;
	char *search_iter;
	size_t files_relabeled = 0;

	if (getenv("PAXRELABEL_DEBUG") != NULL)
		debug = true;

	search_iter = strtok_r(search_path, ":", &search_saveptr);

	do
	{
		if (!process_directory(search_iter, &files_relabeled))
		{
			err(EXIT_FAILURE, "error while relabeling %s", search_iter);
		}

		search_iter = strtok_r(NULL, ":", &search_saveptr);
	} while (search_iter != NULL);

	fprintf(stderr, "paxrelabel: relabeled %zu file%s\n", files_relabeled, files_relabeled != 1 ? "s" : "");

	return EXIT_SUCCESS;
}
